{% extends "admin/submit_line.html" %}
{% load i18n admin_urls %}

{% block submit-row %}
{% if show_publish %}<input type="submit" value="{% translate 'Publish' %}" class="default" name="_publish" />{% endif %}
{{ block.super }}
{% endblock %}
